#!/usr/bin/env python3
"""Script to build mtgjson testdata."""

import collections
import json
import os

from mtg_ssm.mtgjson import downloader

TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')
DATA_DIR = os.path.join(TEST_DATA_DIR, 'source_data')
TARGET_MTGJSON_FILE = os.path.join(TEST_DATA_DIR, 'AllSets_testdata.json')

INCLUDED = {
    'LEA': {'Air Elemental', 'Dark Ritual', 'Forest'},
    'FEM': {'Thallid'},
    'S00': {'Rhox'},
    'ICE': {'Dark Ritual', 'Forest', 'Snow-Covered Forest'},
    'VMA': {'Academy Elite'},
    'pMGD': {'Black Sun\'s Zenith'},
    'HML': {'Cemetery Gate'},
    'ISD': {
        'Abattoir Ghoul', 'Delver of Secrets', 'Insectile Aberration',
        'Forest'},
    'ARC': {'All in Good Time', 'Leonin Abunas'},
    'HOP': {'Academy at Tolaria West', 'Akroma\'s Vengeance', 'Dark Ritual'},
    'PC2': {'Akoum', 'Armored Griffin', 'Chaotic Æther', 'Stairs to Infinity'},
    'MMA': {'Thallid'},
    'pMEI': {'Arena'},
    'PLS': {'Ertai, the Corrupted'},
}


def main():
    """Filter source mtgjson data and dump testdata."""
    print('Fetching mtgjson data.')
    downloader.fetch_mtgjson(DATA_DIR)
    print('Reading mtgjson data.')
    mtg_data = downloader.read_mtgjson(DATA_DIR)

    print('Generating testdata.')
    testdata = collections.OrderedDict()
    for setcode, setdata in mtg_data.items():
        if setcode not in INCLUDED:
            continue
        testsetdata = collections.OrderedDict(
            (k, v) for k, v in setdata.items() if k != 'cards')
        testsetdata['cards'] = [
            c for c in setdata['cards'] if c['name'] in INCLUDED[setcode]]
        testdata[setcode] = testsetdata

    print('Writing testdata.')
    with open(TARGET_MTGJSON_FILE, 'w') as mtgjson_testfile:
        json.dump(
            testdata, mtgjson_testfile, ensure_ascii=False, indent=2,
            sort_keys=False)
    print('Done')


if __name__ == '__main__':
    main()



"""Tests for mtg_ssm.mtgcsv"""

from mtg_ssm.db import models
from mtg_ssm.mtgjson import mtgjson
from mtg_ssm.serialization import mtgcsv

from tests.mtgjson import mtgjson_testcase
from tests.db import sqlite_testcase


class MtgCsvTest(
        sqlite_testcase.SqliteTestCase, mtgjson_testcase.MtgJsonTestCase):

    def setUp(self):
        super().setUp()
        connection = self.engine.connect()
        models.Base.metadata.create_all(connection)
        connection.close()

    def test_get_header(self):
        # Execute
        header = mtgcsv.header()

        # Verify
        expected = [
            'set', 'name', 'number', 'multiverseid', 'mtgjid', 'copies',
            'foils']
        self.assertEqual(expected, header)

    def test_dump_rows(self):
        # Setup
        mtg_data = {
            k: v for k, v in self.mtg_data.items()
            if k in ['ICE', 'S00', 'MMA']}
        mtgjson.update_models(self.session, mtg_data, False)
        self.session.commit()
        forest1 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2746).first()
        forest2 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2747).first()
        forest3 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2748).first()
        forest1.counts['copies'] = 1
        forest2.counts['foils'] = 2
        forest3.counts['copies'] = 3
        forest3.counts['foils'] = 4
        self.session.commit()

        # Execute
        rows = list(mtgcsv.dump_rows(self.session))

        # Verify
        # pylint: disable=line-too-long
        expected = [
            {'set': 'ICE', 'name': 'Dark Ritual', 'multiverseid': 2444, 'mtgjid': '2fab0ea29e3bbe8bfbc981a4c8163f3e7d267853', 'number': None},
            {'set': 'ICE', 'name': 'Forest', 'multiverseid': 2746, 'mtgjid': '676a1f5b64dc03bbb3876840c3ff2ba2c16f99cb', 'number': None, 'copies': 1},
            {'set': 'ICE', 'name': 'Forest', 'multiverseid': 2747, 'mtgjid': 'd0a4414893bc2f9bd3beea2f8f2693635ef926a4', 'number': None, 'foils': 2},
            {'set': 'ICE', 'name': 'Forest', 'multiverseid': 2748, 'mtgjid': 'c78d2da78c68c558b1adc734b3f164e885407ffc', 'number': None, 'copies': 3, 'foils': 4},
            {'set': 'ICE', 'name': 'Snow-Covered Forest', 'multiverseid': 2749, 'mtgjid': '5e9f08498a9343b1954103e493da2586be0fe394', 'number': None},
            {'set': 'S00', 'name': 'Rhox', 'multiverseid': None, 'mtgjid': '536d407161fa03eddee7da0e823c2042a8fa0262', 'number': None},
            {'set': 'MMA', 'name': 'Thallid', 'multiverseid': 370352, 'mtgjid': 'fc46a4b72d216117a352f59217a84d0baeaaacb7', 'number': '167'},
        ]
        # pylint: enable=line-too-long
        self.assertEqual(expected, rows)

    def test_int_or_none(self):
        self.assertEqual(None, mtgcsv.int_or_none(None))
        self.assertEqual(None, mtgcsv.int_or_none(''))
        self.assertEqual(1, mtgcsv.int_or_none('1'))
        self.assertEqual(1, mtgcsv.int_or_none(1))
        self.assertEqual(0, mtgcsv.int_or_none('0'))
        self.assertEqual(0, mtgcsv.int_or_none(0))

    def test_process_row_dict(self):
        # Setup
        # pylint: disable=line-too-long
        rows = [
            {'set': 'FOO', 'name': 'Thing', 'multiverseid': '27', 'number': '52a', 'copies': '1', 'mtgjid': 'ABC'},
            {'set': 'BAR', 'name': 'Another Thing', 'multiverseid': '', 'number': '57', 'foils': '0', 'copies': '', 'mtgjid': 'DEF'},
        ]
        # pylint: enable=line-too-long

        # Execute
        card_dicts = [mtgcsv.process_row_dict(r) for r in rows]

        # Verify
        # pylint: disable=line-too-long
        expected = [
            {'set': 'FOO', 'name': 'Thing', 'multiverseid': 27, 'number': '52a', 'id': 'ABC', 'copies': 1},
            {'set': 'BAR', 'name': 'Another Thing', 'multiverseid': None, 'number': '57', 'id': 'DEF', 'foils': 0, 'copies': None},
        ]
        # pylint: disable=line-too-long
        self.assertEqual(expected, card_dicts)

    def test_read_row_counts(self):
        # Setup
        mtgjson.update_models(self.session, self.mtg_data, False)
        self.session.commit()
        forest1 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2746).first()
        forest2 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2747).first()
        forest3 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2748).first()
        forest4 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2749).first()
        forest4.counts['copies'] = 2
        forest4.counts['foils'] = 3
        self.session.commit()
        # pylint: disable=line-too-long
        rows = [
            {'set': 'ICE', 'name': 'Forest', 'multiverseid': '2746', 'mtgjid': '676a1f5b64dc03bbb3876840c3ff2ba2c16f99cb', 'number': '', 'copies': '1'},
            {'set': 'ICE', 'name': 'Forest', 'multiverseid': '2747', 'mtgjid': 'd0a4414893bc2f9bd3beea2f8f2693635ef926a4', 'number': '', 'foils': '2'},
            {'set': 'ICE', 'name': 'Forest', 'multiverseid': '2748', 'mtgjid': 'c78d2da78c68c558b1adc734b3f164e885407ffc', 'number': '', 'copies': '3', 'foils': '4'},
            {'set': 'ICE', 'name': 'Snow-Covered Forest', 'multiverseid': '2749', 'mtgjid': '5e9f08498a9343b1954103e493da2586be0fe394', 'number': '', 'copies': '', 'foils': ''},
        ]
        # pylint: enable=line-too-long

        # Execute
        mtgcsv.read_row_counts(self.session, rows)
        self.session.commit()

        # Verify
        self.assertEqual({'copies': 1}, forest1.counts)
        self.assertEqual({'foils': 2}, forest2.counts)
        self.assertEqual({'copies': 3, 'foils': 4}, forest3.counts)
        self.assertFalse(forest4.counts)



"""Test helper providing access to test data."""

import json
import os
import unittest

MTGJSON_FILE = os.path.join(
    os.path.dirname(__file__), 'data', 'AllSets_testdata.json')


class MtgJsonTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.mtg_datafile = os.path.abspath(MTGJSON_FILE)
        with open(cls.mtg_datafile, 'r') as testdatafile:
            cls.mtg_data = json.load(testdatafile)



"""Tests for mtg_ssm.mtgcsv"""

from mtg_ssm.db import models
from mtg_ssm.mtgjson import mtgjson
from mtg_ssm.serialization import mtgdict

from tests.mtgjson import mtgjson_testcase
from tests.db import sqlite_testcase


class MtgDictTest(
        sqlite_testcase.SqliteTestCase, mtgjson_testcase.MtgJsonTestCase):

    def setUp(self):
        super().setUp()
        connection = self.engine.connect()
        models.Base.metadata.create_all(connection)
        connection.close()

    def test_get_printing_not_found(self):
        # Setup
        id_to_print = {'ID': 'printing'}
        set_name_num_mv_to_prints = {('SET', 'NAME', 'NUM', 'MV'): ['printing']}
        set_name_mv_to_prints = {('SET', 'NAME', 'MV'): ['printing']}
        set_name_num_to_prints = {('SET', 'NAME', 'NUM'): ['printing']}

        card_dict = {}

        # Execute
        printing = mtgdict.get_printing(
            card_dict, id_to_print, set_name_num_mv_to_prints,
            set_name_mv_to_prints, set_name_num_to_prints)

        # Verify
        self.assertIsNone(printing)

    def test_get_printing_set_name_num(self):
        # Setup
        id_to_print = {'ID': 'printing'}
        set_name_num_mv_to_prints = {('SET', 'NAME', 'NUM', 'MV'): ['printing']}
        set_name_mv_to_prints = {('SET', 'NAME', 'MV'): ['printing']}
        set_name_num_to_prints = {('SET', 'NAME', 'NUM'): ['printing']}

        card_dict = {'set': 'SET', 'name': 'NAME', 'number': 'NUM'}

        # Execute
        printing = mtgdict.get_printing(
            card_dict, id_to_print, set_name_num_mv_to_prints,
            set_name_mv_to_prints, set_name_num_to_prints)

        # Verify
        self.assertEqual('printing', printing)

    def test_get_printing_set_name_num_duplicate(self):
        # Setup
        id_to_print = {'ID': 'printing'}
        set_name_num_mv_to_prints = {('SET', 'NAME', 'NUM', 'MV'): ['printing']}
        set_name_mv_to_prints = {('SET', 'NAME', 'MV'): ['printing']}
        set_name_num_to_prints = {('SET', 'NAME', 'NUM'): ['printing', 'p2']}

        card_dict = {'set': 'SET', 'name': 'NAME', 'number': 'NUM'}

        # Execute
        printing = mtgdict.get_printing(
            card_dict, id_to_print, set_name_num_mv_to_prints,
            set_name_mv_to_prints, set_name_num_to_prints)

        # Verify
        self.assertIsNone(printing)

    def test_get_printing_set_name_mv(self):
        # Setup
        id_to_print = {'ID': 'printing'}
        set_name_num_mv_to_prints = {('SET', 'NAME', 'NUM', 'MV'): ['printing']}
        set_name_mv_to_prints = {('SET', 'NAME', 'MV'): ['printing']}
        set_name_num_to_prints = {('SET', 'NAME', 'NUM'): ['printing']}

        card_dict = {'set': 'SET', 'name': 'NAME', 'multiverseid': 'MV'}

        # Execute
        printing = mtgdict.get_printing(
            card_dict, id_to_print, set_name_num_mv_to_prints,
            set_name_mv_to_prints, set_name_num_to_prints)

        # Verify
        self.assertEqual('printing', printing)

    def test_get_printing_set_name_mv_duplicate(self):
        # Setup
        id_to_print = {'ID': 'printing'}
        set_name_num_mv_to_prints = {('SET', 'NAME', 'NUM', 'MV'): ['printing']}
        set_name_mv_to_prints = {('SET', 'NAME', 'MV'): ['printing', 'p2']}
        set_name_num_to_prints = {('SET', 'NAME', 'NUM'): ['printing']}

        card_dict = {'set': 'SET', 'name': 'NAME', 'multiverseid': 'MV'}

        # Execute
        printing = mtgdict.get_printing(
            card_dict, id_to_print, set_name_num_mv_to_prints,
            set_name_mv_to_prints, set_name_num_to_prints)

        # Verify
        self.assertIsNone(printing)

    def test_get_printing_set_name_num_mv(self):
        # Setup
        id_to_print = {'ID': 'printing'}
        set_name_num_mv_to_prints = {('SET', 'NAME', 'NUM', 'MV'): ['printing']}
        set_name_mv_to_prints = {
            ('SET', 'NAME', 'MV'): ['printing', 'p2', 'p3']}
        set_name_num_to_prints = {
            ('SET', 'NAME', 'NUM'): ['printing', 'p2', 'p3']}

        card_dict = {
            'set': 'SET', 'name': 'NAME', 'number': 'NUM', 'multiverseid': 'MV'}

        # Execute
        printing = mtgdict.get_printing(
            card_dict, id_to_print, set_name_num_mv_to_prints,
            set_name_mv_to_prints, set_name_num_to_prints)

        # Verify
        self.assertEqual('printing', printing)

    def test_get_printing_set_name_num_mv_duplicate(self):
        # Setup
        id_to_print = {'ID': 'printing'}
        set_name_num_mv_to_prints = {
            ('SET', 'NAME', 'NUM', 'MV'): ['printing', 'p2']}
        set_name_mv_to_prints = {
            ('SET', 'NAME', 'MV'): ['printing', 'p2', 'p3']}
        set_name_num_to_prints = {
            ('SET', 'NAME', 'NUM'): ['printing', 'p2', 'p3']}

        card_dict = {
            'set': 'SET', 'name': 'NAME', 'number': 'NUM', 'multiverseid': 'MV'}

        # Execute
        printing = mtgdict.get_printing(
            card_dict, id_to_print, set_name_num_mv_to_prints,
            set_name_mv_to_prints, set_name_num_to_prints)

        # Verify
        self.assertIsNone(printing)

    def test_get_printing_none_set(self):
        # Setup
        id_to_print = {'ID': 'printing'}
        set_name_num_mv_to_prints = {(None, 'NAME', 'NUM', 'MV'): ['printing']}
        set_name_mv_to_prints = {(None, 'NAME', 'MV'): ['printing']}
        set_name_num_to_prints = {(None, 'NAME', 'NUM'): ['printing']}

        card_dict = {'name': 'NAME', 'number': 'NUM', 'multiverseid': 'MV'}

        # Execute
        printing = mtgdict.get_printing(
            card_dict, id_to_print, set_name_num_mv_to_prints,
            set_name_mv_to_prints, set_name_num_to_prints)

        # Verify
        self.assertIsNone(printing)

    def test_get_printing_none_name(self):
        # Setup
        id_to_print = {'ID': 'printing'}
        set_name_num_mv_to_prints = {('SET', None, 'NUM', 'MV'): ['printing']}
        set_name_mv_to_prints = {('SET', None, 'MV'): ['printing']}
        set_name_num_to_prints = {('SET', None, 'NUM'): ['printing']}

        card_dict = {'set': 'SET', 'number': 'NUM', 'multiverseid': 'MV'}

        # Execute
        printing = mtgdict.get_printing(
            card_dict, id_to_print, set_name_num_mv_to_prints,
            set_name_mv_to_prints, set_name_num_to_prints)

        # Verify
        self.assertIsNone(printing)

    def test_get_printing_by_id(self):
        # Setup
        id_to_print = {'ID': 'printing'}
        set_name_num_mv_to_prints = {('SET', 'NAME', 'NUM', 'MV'): ['printing']}
        set_name_mv_to_prints = {('SET', 'NAME', 'MV'): ['printing']}
        set_name_num_to_prints = {('SET', 'NAME', 'NUM'): ['printing']}

        card_dict = {'id': 'ID'}

        # Execute
        printing = mtgdict.get_printing(
            card_dict, id_to_print, set_name_num_mv_to_prints,
            set_name_mv_to_prints, set_name_num_to_prints)

        # Verify
        self.assertEqual('printing', printing)

    def test_load_counts(self):
        # Setup
        mtgjson.update_models(self.session, self.mtg_data, False)
        self.session.commit()

        forest1 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2746).first()
        forest2 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2747).first()
        forest3 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2748).first()
        forest4 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2749).first()
        forest4.counts['copies'] = 2
        forest4.counts['foils'] = 3
        self.session.commit()
        # pylint: disable=line-too-long
        card_dicts = [
            {'set': 'ICE', 'name': 'Forest', 'multiverseid': 2746, 'number': None, 'copies': 1},
            {'set': 'ICE', 'name': 'Forest', 'multiverseid': 2747, 'number': None, 'foils': 2},
            {'set': 'ICE', 'name': 'Forest', 'multiverseid': 2748, 'number': None, 'copies': 3, 'foils': 4},
            {'set': 'ICE', 'name': 'Snow-Covered Forest', 'multiverseid': 2749, 'number': None, 'copies': None, 'foils': None},
        ]
        # pylint: enable=line-too-long

        # Execute
        mtgdict.load_counts(self.session, card_dicts)
        self.session.commit()

        # Verify
        self.assertEqual({'copies': 1}, forest1.counts)
        self.assertEqual({'foils': 2}, forest2.counts)
        self.assertEqual({'copies': 3, 'foils': 4}, forest3.counts)
        self.assertFalse(forest4.counts)



"""Tests for mtg_ssm.mtgxlsx"""

import collections
import datetime
from unittest import mock

import openpyxl

from mtg_ssm.db import models
from mtg_ssm.mtgjson import mtgjson
from mtg_ssm.serialization import mtgxlsx

from tests.mtgjson import mtgjson_testcase
from tests.db import sqlite_testcase


class MtgXlsxTest(
        sqlite_testcase.SqliteTestCase, mtgjson_testcase.MtgJsonTestCase):

    def setUp(self):
        super().setUp()
        connection = self.engine.connect()
        models.Base.metadata.create_all(connection)
        connection.close()

    def test_create_sets_sheet(self):
        # Setup
        mtgjson.update_models(self.session, self.mtg_data, True)
        self.session.commit()
        card_sets = self.session.query(models.CardSet) \
            .order_by(models.CardSet.release_date).all()
        book = openpyxl.Workbook()
        sheet = book.create_sheet()

        # Execute
        mtgxlsx.create_sets_sheet(sheet, card_sets)

        # Verify
        rows = [[cell.value for cell in row] for row in sheet.rows]
        # pylint: disable=line-too-long
        expected = [
            ['code', 'name', 'release', 'block', 'type', 'cards', 'unique', 'playsets', 'count'],
            ['Total', None, None, None, None, '=SUM(F3:F65535)', '=SUM(G3:G65535)', '=SUM(H3:H65535)', '=SUM(I3:I63353)'],
            ['LEA', 'Limited Edition Alpha', datetime.datetime(1993, 8, 5), None, 'core', 4, '=COUNTIF(\'LEA\'!A:A,">0")', '=COUNTIF(\'LEA\'!A:A,">=4")', "=SUM('LEA'!A:A)"],
            ['FEM', 'Fallen Empires', datetime.datetime(1994, 11, 1), None, 'expansion', 4, '=COUNTIF(\'FEM\'!A:A,">0")', '=COUNTIF(\'FEM\'!A:A,">=4")', "=SUM('FEM'!A:A)"],
            ['pMEI', 'Media Inserts', datetime.datetime(1995, 1, 1), None, 'promo', 1, '=COUNTIF(\'pMEI\'!A:A,">0")', '=COUNTIF(\'pMEI\'!A:A,">=4")', "=SUM('pMEI'!A:A)"],
            ['ICE', 'Ice Age', datetime.datetime(1995, 6, 1), 'Ice Age', 'expansion', 5, '=COUNTIF(\'ICE\'!A:A,">0")', '=COUNTIF(\'ICE\'!A:A,">=4")', "=SUM('ICE'!A:A)"],
            ['HML', 'Homelands', datetime.datetime(1995, 10, 1), None, 'expansion', 2, '=COUNTIF(\'HML\'!A:A,">0")', '=COUNTIF(\'HML\'!A:A,">=4")', "=SUM('HML'!A:A)"],
            ['S00', 'Starter 2000', datetime.datetime(2000, 4, 1), None, 'starter', 1, '=COUNTIF(\'S00\'!A:A,">0")', '=COUNTIF(\'S00\'!A:A,">=4")', "=SUM('S00'!A:A)"],
            ['PLS', 'Planeshift', datetime.datetime(2001, 2, 5), 'Invasion', 'expansion', 2, '=COUNTIF(\'PLS\'!A:A,">0")', '=COUNTIF(\'PLS\'!A:A,">=4")', "=SUM('PLS'!A:A)"],
            ['pMGD', 'Magic Game Day', datetime.datetime(2007, 7, 14), None, 'promo', 1, '=COUNTIF(\'pMGD\'!A:A,">0")', '=COUNTIF(\'pMGD\'!A:A,">=4")', "=SUM('pMGD'!A:A)"],
            ['HOP', 'Planechase', datetime.datetime(2009, 9, 4), None, 'planechase', 3, '=COUNTIF(\'HOP\'!A:A,">0")', '=COUNTIF(\'HOP\'!A:A,">=4")', "=SUM('HOP'!A:A)"],
            ['ARC', 'Archenemy', datetime.datetime(2010, 6, 18), None, 'archenemy', 2, '=COUNTIF(\'ARC\'!A:A,">0")', '=COUNTIF(\'ARC\'!A:A,">=4")', "=SUM('ARC'!A:A)"],
            ['ISD', 'Innistrad', datetime.datetime(2011, 9, 30), 'Innistrad', 'expansion', 6, '=COUNTIF(\'ISD\'!A:A,">0")', '=COUNTIF(\'ISD\'!A:A,">=4")', "=SUM('ISD'!A:A)"],
            ['PC2', 'Planechase 2012 Edition', datetime.datetime(2012, 6, 1), None, 'planechase', 4, '=COUNTIF(\'PC2\'!A:A,">0")', '=COUNTIF(\'PC2\'!A:A,">=4")', "=SUM('PC2'!A:A)"],
            ['MMA', 'Modern Masters', datetime.datetime(2013, 6, 7, 0, 0), None, 'reprint', 1, '=COUNTIF(\'MMA\'!A:A,">0")', '=COUNTIF(\'MMA\'!A:A,">=4")', "=SUM('MMA'!A:A)"],
            ['VMA', 'Vintage Masters', datetime.datetime(2014, 6, 16), None, 'masters', 1, '=COUNTIF(\'VMA\'!A:A,">0")', '=COUNTIF(\'VMA\'!A:A,">=4")', "=SUM('VMA'!A:A)"],
        ]
        # pylint: enable=line-too-long
        self.assertEqual(expected, rows)

    def test_split_into_consecutives(self):
        inlist = []
        expected = []
        self.assertEqual(expected, mtgxlsx.split_into_consecutives(inlist))

        inlist = [1]
        expected = [[1]]
        self.assertEqual(expected, mtgxlsx.split_into_consecutives(inlist))

        inlist = [1, 2, 3]
        expected = [[1, 2, 3]]
        self.assertEqual(expected, mtgxlsx.split_into_consecutives(inlist))

        inlist = [1, 2, 3, 5, 6, 8, 9, 10, 11]
        expected = [[1, 2, 3], [5, 6], [8, 9, 10, 11]]
        self.assertEqual(expected, mtgxlsx.split_into_consecutives(inlist))

        inlist = [4, 3, 7, 1]
        expected = [[1], [3, 4], [7]]
        self.assertEqual(expected, mtgxlsx.split_into_consecutives(inlist))

    def test_create_haveref_sum(self):
        setcode = 'ABC'

        rownums = [1]
        expected = "'ABC'!A1"
        self.assertEqual(expected, mtgxlsx.create_haveref_sum(setcode, rownums))

        rownums = [3, 4, 5, 6]
        expected = "SUM('ABC'!A3:A6)"
        self.assertEqual(expected, mtgxlsx.create_haveref_sum(setcode, rownums))

        rownums = [3, 4, 5, 8, 10]
        expected = "SUM('ABC'!A3:A5)+'ABC'!A8+'ABC'!A10"
        self.assertEqual(expected, mtgxlsx.create_haveref_sum(setcode, rownums))

    def test_get_other_print_refs_multiple_sets(self):
        # Setup
        mtgjson.update_models(self.session, self.mtg_data, True)
        self.session.commit()
        dark_rits = self.session.query(models.CardPrinting).filter(
            models.CardPrinting.card.has(name='Dark Ritual')).all()
        name_to_prints = {'Dark Ritual': dark_rits}
        lea_dark_rit = self.session.query(
            models.CardPrinting).filter_by(multiverseid=54).first()

        # Execute
        print_refs = mtgxlsx.get_other_print_references(
            lea_dark_rit, name_to_prints)

        # Verify
        expected = (
            '=IF(\'ICE\'!A2>0,"ICE: "&\'ICE\'!A2&", ","")'
            '&IF(\'HOP\'!A4>0,"HOP: "&\'HOP\'!A4&", ","")')
        self.assertEqual(expected, print_refs)

    def test_get_other_print_refs_multiple_variants(self):
        # Setup
        mtgjson.update_models(self.session, self.mtg_data, True)
        self.session.commit()
        thallids = self.session.query(models.CardPrinting).filter(
            models.CardPrinting.card.has(name='Thallid')).all()
        name_to_prints = {'Thallid': thallids}
        mma_thallid = self.session.query(
            models.CardPrinting).filter_by(multiverseid=370352).first()

        # Execute
        print_refs = mtgxlsx.get_other_print_references(
            mma_thallid, name_to_prints)

        # Verify
        expected = (
            '=IF(SUM(\'FEM\'!A2:A5)>0,"FEM: "&SUM(\'FEM\'!A2:A5)&", ","")')
        self.assertEqual(expected, print_refs)

    def test_get_other_print_references_basic_land(self):
        # Setup
        mtgjson.update_models(self.session, self.mtg_data, True)
        self.session.commit()
        forests = self.session.query(models.CardPrinting).filter(
            models.CardPrinting.card.has(name='Forest')).all()
        name_to_prints = {'Forest': forests}
        lea_forest = self.session.query(
            models.CardPrinting).filter_by(multiverseid=288).first()

        # Execute
        print_refs = mtgxlsx.get_other_print_references(
            lea_forest, name_to_prints)

        # Verify
        self.assertIsNone(print_refs)

    def test_create_cards_sheet(self):
        # Setup
        mtgjson.update_models(self.session, self.mtg_data, True)
        self.session.commit()
        forest1 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2746).first()
        forest2 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2747).first()
        forest3 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2748).first()
        forest1.counts['copies'] = 1
        forest2.counts['foils'] = 2
        forest3.counts['copies'] = 3
        forest3.counts['foils'] = 4
        self.session.commit()
        printings = self.session.query(models.CardPrinting)
        name_to_prints = collections.defaultdict(list)
        for printing in printings:
            name_to_prints[printing.card.name].append(printing)
        ice_age = self.session.query(models.CardSet).filter_by(code='ICE').one()
        book = openpyxl.Workbook()
        sheet = book.create_sheet()

        # Execute
        mtgxlsx.create_cards_sheet(sheet, ice_age, name_to_prints)

        # Verify
        rows = [[cell.value for cell in row] for row in sheet.rows]
        # pylint: disable=line-too-long
        expected = [
            ['have', 'name', 'id', 'multiverseid', 'number', 'artist', 'copies', 'foils', 'others'],
            ['=G2+H2', 'Dark Ritual', '2fab0ea29e3bbe8bfbc981a4c8163f3e7d267853', 2444, None, 'Justin Hampton', None, None, mock.ANY],
            ['=G3+H3', 'Forest', '676a1f5b64dc03bbb3876840c3ff2ba2c16f99cb', 2746, None, 'Pat Morrissey', 1, None, mock.ANY],
            ['=G4+H4', 'Forest', 'd0a4414893bc2f9bd3beea2f8f2693635ef926a4', 2747, None, 'Pat Morrissey', None, 2, mock.ANY],
            ['=G5+H5', 'Forest', 'c78d2da78c68c558b1adc734b3f164e885407ffc', 2748, None, 'Pat Morrissey', 3, 4, mock.ANY],
            ['=G6+H6', 'Snow-Covered Forest', '5e9f08498a9343b1954103e493da2586be0fe394', 2749, None, 'Pat Morrissey', None, None, mock.ANY],
        ]
        # pylint: enable=line-too-long
        self.assertEqual(expected, rows)

    def test_dump_workbook(self):
        # Setup
        mtgjson.update_models(self.session, self.mtg_data, True)
        self.session.commit()

        # Execute
        book = mtgxlsx.dump_workbook(self.session)

        # Verify
        expected = [
            'Sets',
            'LEA',
            'FEM',
            'pMEI',
            'ICE',
            'HML',
            'S00',
            'PLS',
            'pMGD',
            'HOP',
            'ARC',
            'ISD',
            'PC2',
            'MMA',
            'VMA',
        ]
        self.assertEqual(expected, book.sheetnames)

    def test_worksheet_row_reader(self):
        # Setup
        book = openpyxl.Workbook()
        sheet = book.create_sheet()
        sheet.title = 'ICE'
        sheet_data = [
            ['name', 'multiverseid', 'number', 'artist', 'copies', 'foils'],
            ['Forest', 2746, None, 'Pat Morrissey', 1, None],
            ['Forest', 2747, None, 'Pat Morrissey', None, 2],
            ['Forest', 2748, None, 'Pat Morrissey', 3, 4],
            ['Snow-Covered Forest', 2749, None, 'Pat Morrissey', None, None],
        ]
        for row in sheet_data:
            sheet.append(row)

        # Execute
        row_dicts = mtgxlsx.worksheet_row_reader(sheet)

        # Verify
        expected = [
            # pylint: disable=line-too-long
            {'name': 'Forest', 'artist': 'Pat Morrissey', 'multiverseid': 2746, 'number': None, 'copies': 1, 'foils': None},
            {'name': 'Forest', 'artist': 'Pat Morrissey', 'multiverseid': 2747, 'number': None, 'copies': None, 'foils': 2},
            {'name': 'Forest', 'artist': 'Pat Morrissey', 'multiverseid': 2748, 'number': None, 'copies': 3, 'foils': 4},
            {'name': 'Snow-Covered Forest', 'artist': 'Pat Morrissey', 'multiverseid': 2749, 'number': None, 'copies': None, 'foils': None},
            # pylint: enable=line-too-long
        ]
        self.assertEqual(expected, list(row_dicts))

    def test_workbook_row_reader(self):
        # Setup
        book = openpyxl.Workbook()
        sheet = book.create_sheet()
        sheet.title = 'ICE'
        sheet_data = [
            ['name', 'multiverseid', 'number', 'artist', 'copies', 'foils'],
            ['Forest', 2746, None, 'Pat Morrissey', 1, None],
            ['Forest', 2747, None, 'Pat Morrissey', None, 2],
            ['Forest', 2748, None, 'Pat Morrissey', 3, 4],
            ['Snow-Covered Forest', 2749, None, 'Pat Morrissey', None, None],
        ]
        for row in sheet_data:
            sheet.append(row)
        known_sets = {'ICE'}

        # Execute
        row_dicts = mtgxlsx.workbook_row_reader(book, known_sets)

        # Verify
        expected = [
            # pylint: disable=line-too-long
            {'set': 'ICE', 'name': 'Forest', 'artist': 'Pat Morrissey', 'multiverseid': 2746, 'number': None, 'copies': 1, 'foils': None},
            {'set': 'ICE', 'name': 'Forest', 'artist': 'Pat Morrissey', 'multiverseid': 2747, 'number': None, 'copies': None, 'foils': 2},
            {'set': 'ICE', 'name': 'Forest', 'artist': 'Pat Morrissey', 'multiverseid': 2748, 'number': None, 'copies': 3, 'foils': 4},
            {'set': 'ICE', 'name': 'Snow-Covered Forest', 'artist': 'Pat Morrissey', 'multiverseid': 2749, 'number': None, 'copies': None, 'foils': None},
            # pylint: enable=line-too-long
        ]
        self.assertEqual(expected, list(row_dicts))

    def test_workbook_row_reader_invalid_set(self):
        # Setup
        book = openpyxl.Workbook()
        sheet = book.create_sheet()
        sheet.title = 'garbage'
        sheet.append(['jabberwocky', 'gobbledy', 'goop'])
        known_sets = {'ICE'}

        # Execute
        row_dicts = mtgxlsx.workbook_row_reader(book, known_sets)

        # Verify
        self.assertFalse(list(row_dicts))

    def test_read_workbook_counts(self):
        # Setup
        mtgjson.update_models(self.session, self.mtg_data, True)
        self.session.commit()
        forest1 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2746).first()
        forest2 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2747).first()
        forest3 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2748).first()
        forest4 = self.session.query(
            models.CardPrinting).filter_by(multiverseid=2749).first()
        forest4.counts['copies'] = 2
        forest4.counts['foils'] = 3
        self.session.commit()
        book = openpyxl.Workbook()
        sets_sheet = book.create_sheet()
        sets_sheet.title = 'Sets'
        sets_sheet.append(['jabberwocky', 'gobbledy', 'goop'])

        cards_sheet = book.create_sheet()
        cards_sheet.title = 'ICE'
        sheet_data = [
            ['name', 'multiverseid', 'number', 'artist', 'copies', 'foils'],
            ['Forest', 2746, None, 'Pat Morrissey', 1, None],
            ['Forest', 2747, None, 'Pat Morrissey', None, 2],
            ['Forest', 2748, None, 'Pat Morrissey', 3, 4],
            ['Snow-Covered Forest', 2749, None, 'Pat Morrissey', None, None],
        ]
        for row in sheet_data:
            cards_sheet.append(row)

        # Execute
        mtgxlsx.read_workbook_counts(self.session, book)
        self.session.commit()

        # Verify
        self.assertEqual({'copies': 1}, forest1.counts)
        self.assertEqual({'foils': 2}, forest2.counts)
        self.assertEqual({'copies': 3, 'foils': 4}, forest3.counts)
        self.assertFalse(forest4.counts)



"""Tests for mtg_ssm.util"""

import enum

import sqlalchemy as sqla
import sqlalchemy.ext.declarative as sqld

from mtg_ssm.db import util

from tests.db import sqlite_testcase


class MyEnum(enum.IntEnum):
    """Simple enum to test SqlEnumType."""
    one = 1
    two = 2


class Model(sqld.declarative_base()):
    """Simple declarative model to test enum handling."""
    __tablename__ = 'test'

    id = sqla.Column(sqla.Integer, primary_key=True)
    enum_col = sqla.Column(util.SqlEnumType(MyEnum))


class SqlEnumTest(sqlite_testcase.SqliteTestCase):

    def setUp(self):
        super().setUp()
        connection = self.engine.connect()
        Model.metadata.create_all(connection)
        connection.close()

    def test_store_retrieve(self):
        # Store
        self.session.add(Model(id=1, enum_col=MyEnum.one))
        self.session.add(Model(id=2, enum_col=MyEnum.two))
        self.session.commit()

        # Retrieve
        objs = self.session.query(Model).all()
        obj_dicts = [{'id': o.id, 'enum_col': o.enum_col} for o in objs]
        expected = [
            {'id': 1, 'enum_col': MyEnum.one},
            {'id': 2, 'enum_col': MyEnum.two},
        ]
        self.assertEqual(expected, obj_dicts)

    def test_db_strings(self):
        # Setup
        self.session.add(Model(id=1, enum_col=MyEnum.one))
        self.session.add(Model(id=2, enum_col=MyEnum.two))
        self.session.commit()

        # Verify
        connection = self.engine.connect()
        rows = connection.execute('SELECT * FROM test')
        id_to_enum_col = {r['id']: r['enum_col'] for r in rows}
        connection.close()
        expected = {
            1: 'one',
            2: 'two',
        }
        self.assertEqual(expected, id_to_enum_col)



"""Test for mtgjson_testcase."""

from tests.mtgjson import mtgjson_testcase


class MtgJsonTestCaseTest(mtgjson_testcase.MtgJsonTestCase):

    def test_has_mtgjson(self):
        self.assertTrue(self.mtg_data)

    def test_lea(self):
        lea = self.mtg_data['LEA']
        self.assertIsInstance(lea, dict)

    def test_lea_cards(self):
        lea = self.mtg_data['LEA']
        lea_cards = lea['cards']
        self.assertIsInstance(lea_cards, list)

    def test_lea_air_elemental(self):
        [air_elemental] = [
            c for c in self.mtg_data['LEA']['cards']
            if c['id'] == '926234c2fe8863f49220a878346c4c5ca79b6046']
        self.assertIsInstance(air_elemental, dict)
        self.assertEqual('Air Elemental', air_elemental['name'])
        self.assertEqual(94, air_elemental['multiverseid'])

    def test_chaotic_aether(self):
        [chaotic_aether] = [
            c for c in self.mtg_data['PC2']['cards']
            if c['id'] == '5669523e75ffdb436b768d4dd37cb95b82919d51']
        self.assertEqual('Chaotic \u00c6ther', chaotic_aether['name'])



"""Testcase for tests that require a sqlalchemy sqlite engine."""

import unittest

import sqlalchemy
import sqlalchemy.event
import sqlalchemy.orm


def set_sqlite_pragma(dbapi_connection, _connection_record):
    """Listener handler to enable foreign key pragmas on connect."""
    cursor = dbapi_connection.cursor()
    cursor.execute('PRAGMA foreign_keys=ON')
    cursor.close()


class SqliteTestCase(unittest.TestCase):
    """TestCase that creates an sqlite database for testing."""

    def setUp(self):
        super().setUp()
        self.engine = sqlalchemy.create_engine('sqlite://')
        sqlalchemy.event.listen(self.engine, 'connect', set_sqlite_pragma)

        self.sessionmaker = sqlalchemy.orm.sessionmaker(self.engine)
        self.session = self.__session = self.sessionmaker()

    def tearDown(self):
        self.__session.close()
        self.engine.dispose()
        super().tearDown()



"""Tests for mtg_ssm.models"""

import sqlalchemy.exc as sqlx

from mtg_ssm.db import models

from tests.db import sqlite_testcase


class ModelsTest(sqlite_testcase.SqliteTestCase):

    def setUp(self):
        super().setUp()
        connection = self.engine.connect()
        models.Base.metadata.create_all(connection)
        connection.close()

    def test_set_integer_variant(self):
        # Setup
        card_set = models.CardSet(code='F', name='Foo')
        card = models.Card(name='Bar')
        printing = models.CardPrinting(
            id='A', card_name='Bar', set_code='F', set_number='123abc',
            multiverseid=27, artist='Quux')
        self.session.add_all([card_set, card, printing])
        self.session.commit()

        # Verify
        self.assertEqual('123abc', printing.set_number)
        self.assertEqual(123, printing.set_integer)
        self.assertEqual('abc', printing.set_variant)

    def test_set_integer_variant_null(self):
        # Setup
        card_set = models.CardSet(code='F', name='Foo')
        card = models.Card(name='Bar')
        printing = models.CardPrinting(
            id='A', card_name='Bar', set_code='F', set_number=None,
            multiverseid=None, artist='Quux')
        self.session.add_all([card_set, card, printing])
        self.session.commit()

        # Verify
        self.assertEqual(None, printing.set_number)
        self.assertEqual(None, printing.set_integer)
        self.assertEqual(None, printing.set_variant)

    def test_set_integer_variant_nonascii_prefix(self):
        # Setup
        card_set = models.CardSet(code='F', name='Foo')
        card = models.Card(name='Bar')
        printing = models.CardPrinting(
            id='A', card_name='Bar', set_code='F', set_number='★107',
            multiverseid=27, artist='Quux')
        self.session.add_all([card_set, card, printing])
        self.session.commit()

        # Verify
        self.assertEqual('★107', printing.set_number)
        self.assertEqual(107, printing.set_integer)
        self.assertEqual('★', printing.set_variant)

    def test_counts_read(self):
        # Setup
        card_set = models.CardSet(code='F', name='Foo')
        card = models.Card(name='Bar')
        printing = models.CardPrinting(
            id='A', card_name='Bar', set_code='F', set_number='123abc',
            multiverseid=27, artist='Quux')
        count = models.CollectionCount(
            print_id='A', type=models.CountTypes.copies, count=5)
        self.session.add_all([card_set, card, printing, count])
        self.session.commit()

        # Verify
        self.assertEqual(5, printing.collection_counts['copies'].count)
        self.assertEqual(5, printing.counts['copies'])
        with self.assertRaises(KeyError):
            _ = printing.counts['foils']

    def test_counts_write(self):
        # Setup
        card_set = models.CardSet(code='F', name='Foo')
        card = models.Card(name='Bar')
        printing = models.CardPrinting(
            id='A', card_name='Bar', set_code='F', set_number='123abc',
            multiverseid=27, artist='Quux')
        count = models.CollectionCount(
            print_id='A', type=models.CountTypes.copies, count=5)
        self.session.add_all([card_set, card, printing, count])
        self.session.commit()

        # Execute
        printing.counts['copies'] = 2
        printing.counts['foils'] = 7
        self.session.commit()

        # Verify
        counts = self.session.query(models.CollectionCount).all()
        print_type_count = [(c.print_id, c.type, c.count) for c in counts]
        expected = [
            ('A', models.CountTypes.copies, 2),
            ('A', models.CountTypes.foils, 7),
        ]
        self.assertCountEqual(expected, print_type_count)

    def test_invalid_counts_key(self):
        # Setup
        card_set = models.CardSet(code='F', name='Foo')
        card = models.Card(name='Bar')
        printing = models.CardPrinting(
            id='A', card_name='Bar', set_code='F', set_number='123abc',
            multiverseid=27, artist='Quux')
        self.session.add_all([card_set, card, printing])
        self.session.commit()

        # Execute
        with self.assertRaises(AttributeError):
            printing.counts['invalid'] = 12

    def test_invalid_counts_value(self):
        # Setup
        card_set = models.CardSet(code='F', name='Foo')
        card = models.Card(name='Bar')
        printing = models.CardPrinting(
            id='A', card_name='Bar', set_code='F', set_number='123abc',
            multiverseid=27, artist='Quux')
        self.session.add_all([card_set, card, printing])
        self.session.commit()

        # Execute
        with self.assertRaises(sqlx.IntegrityError):
            printing.counts['copies'] = None
            self.session.commit()

    def test_counts_delete(self):
        # Setup
        card_set = models.CardSet(code='F', name='Foo')
        card = models.Card(name='Bar')
        printing = models.CardPrinting(
            id='A', card_name='Bar', set_code='F', set_number='123abc',
            multiverseid=27, artist='Quux')
        count = models.CollectionCount(
            print_id='A', type=models.CountTypes.copies, count=5)
        self.session.add_all([card_set, card, printing, count])
        self.session.commit()

        # Execute
        del printing.counts['copies']
        self.session.commit()

        # Verify
        self.assertEqual({}, printing.counts)
        with self.assertRaises(KeyError):
            _ = printing.counts['copies']



"""Tests for mtg_ssm.mtgjson"""

import datetime

from mtg_ssm.db import models
from mtg_ssm.mtgjson import mtgjson

from tests.mtgjson import mtgjson_testcase
from tests.db import sqlite_testcase


class MtgjsonTest(
        sqlite_testcase.SqliteTestCase, mtgjson_testcase.MtgJsonTestCase):

    def setUp(self):
        super().setUp()
        connection = self.engine.connect()
        models.Base.metadata.create_all(connection)
        connection.close()

    def test_create_set(self):
        # Setup
        set_data = self.mtg_data['ISD']

        # Execute
        ret_card_set = mtgjson.create_set(set_data)
        self.session.merge(ret_card_set)
        self.session.commit()

        # Verify
        [card_set] = self.session.query(models.CardSet).all()
        self.assertEqual('ISD', card_set.code)
        self.assertEqual('Innistrad', card_set.name)
        self.assertEqual(datetime.date(2011, 9, 30), card_set.release_date)
        self.assertEqual('expansion', card_set.type)
        self.assertFalse(card_set.online_only)
        self.assertEqual('Innistrad', card_set.block)
        self.assertEqual([], card_set.printings)

    def test_update_set(self):
        # Setup
        setup_card_set = models.CardSet(code='ISD', name='Bunk')
        self.session.add(setup_card_set)
        self.session.commit()
        set_data = self.mtg_data['ISD']

        # Execute
        ret_card_set = mtgjson.create_set(set_data)
        self.session.merge(ret_card_set)
        self.session.commit()

        # Verify
        [card_set] = self.session.query(models.CardSet).all()
        self.assertEqual('ISD', card_set.code)
        self.assertEqual('Innistrad', card_set.name)
        self.assertEqual(datetime.date(2011, 9, 30), card_set.release_date)
        self.assertEqual('expansion', card_set.type)
        self.assertFalse(card_set.online_only)
        self.assertEqual('Innistrad', card_set.block)
        self.assertEqual([], card_set.printings)

    def test_online_only(self):
        # Execute
        isd_set = mtgjson.create_set(self.mtg_data['ISD'])
        vma_set = mtgjson.create_set(self.mtg_data['VMA'])

        # Verify
        self.assertTrue(vma_set.online_only)
        self.assertFalse(isd_set.online_only)

    def test_create_card(self):
        # Setup
        [card_data] = [
            c for c in self.mtg_data['ISD']['cards']
            if c['id'] == '958ae1416f8f6287115ccd7c5c61f2415a313546']

        # Execute
        ret_card = mtgjson.create_card(card_data)
        self.session.merge(ret_card)
        self.session.commit()

        # Verify
        [card] = self.session.query(models.Card).all()
        self.assertEqual('Abattoir Ghoul', card.name)
        self.assertFalse(card.strict_basic)

    def test_update_card(self):
        # Setup
        setup_card = models.Card(name='Abattoir Ghoul', strict_basic=True)
        self.session.add(setup_card)
        self.session.commit()
        [card_data] = [
            c for c in self.mtg_data['ISD']['cards']
            if c['id'] == '958ae1416f8f6287115ccd7c5c61f2415a313546']

        # Execute
        ret_card = mtgjson.create_card(card_data)
        self.session.merge(ret_card)
        self.session.commit()

        # Verify
        [card] = self.session.query(models.Card).all()
        self.assertEqual('Abattoir Ghoul', card.name)
        self.assertFalse(card.strict_basic)

    def test_strict_basic(self):
        # Execute
        [ag_card_data] = [
            c for c in self.mtg_data['ISD']['cards']
            if c['id'] == '958ae1416f8f6287115ccd7c5c61f2415a313546']
        [f_card_data] = [
            c for c in self.mtg_data['ICE']['cards']
            if c['id'] == '676a1f5b64dc03bbb3876840c3ff2ba2c16f99cb']
        ag_card = mtgjson.create_card(ag_card_data)
        f_card = mtgjson.create_card(f_card_data)

        # Verify
        self.assertEqual('Abattoir Ghoul', ag_card.name)
        self.assertFalse(ag_card.strict_basic)
        self.assertEqual('Forest', f_card.name)
        self.assertTrue(f_card.strict_basic)

    def test_create_printing(self):
        # Setup
        card_set = models.CardSet(code='ISD', name='Innistrad')
        card = models.Card(name='Abattoir Ghoul')
        self.session.add(card_set)
        self.session.add(card)
        self.session.flush()
        [card_data] = [
            c for c in self.mtg_data['ISD']['cards']
            if c['id'] == '958ae1416f8f6287115ccd7c5c61f2415a313546']

        # Execute
        ret_printing = mtgjson.create_printing(card_data, 'ISD')
        self.session.merge(ret_printing)
        self.session.commit()

        # Verify
        [printing] = self.session.query(models.CardPrinting).all()
        self.assertEqual(
            '958ae1416f8f6287115ccd7c5c61f2415a313546', printing.id)
        self.assertEqual('Abattoir Ghoul', printing.card.name)
        self.assertEqual('ISD', printing.set.code)
        self.assertEqual('85', printing.set_number)
        self.assertEqual(222911, printing.multiverseid)
        self.assertEqual('Volkan Baga', printing.artist)

    def test_update_printing(self):
        # Setup
        card_set = models.CardSet(code='ISD', name='Innistrad')
        card = models.Card(name='Abattoir Ghoul')
        setup_printing = models.CardPrinting(
            id='958ae1416f8f6287115ccd7c5c61f2415a313546',
            card_name='Abattoir Ghoul', set_code='ISD', artist='Hokum')
        self.session.add(card_set)
        self.session.add(card)
        self.session.add(setup_printing)
        self.session.commit()
        [card_data] = [
            c for c in self.mtg_data['ISD']['cards']
            if c['id'] == '958ae1416f8f6287115ccd7c5c61f2415a313546']

        # Execute
        ret_printing = mtgjson.create_printing(card_data, 'ISD')
        self.session.merge(ret_printing)
        self.session.commit()

        # Verify
        [printing] = self.session.query(models.CardPrinting).all()
        self.assertEqual(
            '958ae1416f8f6287115ccd7c5c61f2415a313546', printing.id)
        self.assertEqual('Abattoir Ghoul', printing.card.name)
        self.assertEqual('ISD', printing.set.code)
        self.assertEqual('85', printing.set_number)
        self.assertEqual(222911, printing.multiverseid)
        self.assertEqual('Volkan Baga', printing.artist)

    def test_update_models_with_online_only(self):
        # Execute
        mtgjson.update_models(self.session, self.mtg_data, True)
        self.session.commit()

        # Verify
        printings = self.session.query(models.CardPrinting).all()
        set_card_mv_number = [
            (p.set.code, p.card.name, p.multiverseid, p.set_number)
            for p in printings]
        expected = [
            ('LEA', 'Dark Ritual', 54, None),
            ('LEA', 'Air Elemental', 94, None),
            ('LEA', 'Forest', 288, None),
            ('LEA', 'Forest', 289, None),
            ('FEM', 'Thallid', 1924, None),
            ('FEM', 'Thallid', 1926, None),
            ('FEM', 'Thallid', 1927, None),
            ('FEM', 'Thallid', 1925, None),
            ('pMEI', 'Arena', 97042, '1'),
            ('ICE', 'Dark Ritual', 2444, None),
            ('ICE', 'Forest', 2746, None),
            ('ICE', 'Forest', 2747, None),
            ('ICE', 'Forest', 2748, None),
            ('ICE', 'Snow-Covered Forest', 2749, None),
            ('HML', 'Cemetery Gate', 2913, None),
            ('HML', 'Cemetery Gate', 2914, None),
            ('S00', 'Rhox', None, None),
            ('PLS', 'Ertai, the Corrupted', 25614, '107'),
            ('PLS', 'Ertai, the Corrupted', 29292, '★107'),
            ('pMGD', 'Black Sun\'s Zenith', None, '7'),
            ('HOP', 'Academy at Tolaria West', 198073, '1'),
            ('HOP', 'Akroma\'s Vengeance', 205366, '1'),
            ('HOP', 'Dark Ritual', 205422, '24'),
            ('ARC', 'All in Good Time', 212648, '1'),
            ('ARC', 'Leonin Abunas', 220527, '1'),
            ('ISD', 'Abattoir Ghoul', 222911, '85'),
            ('ISD', 'Delver of Secrets', 226749, '51a'),
            ('ISD', 'Insectile Aberration', 226755, '51b'),
            ('ISD', 'Forest', 245247, '262'),
            ('ISD', 'Forest', 245248, '263'),
            ('ISD', 'Forest', 245246, '264'),
            ('PC2', 'Akoum', 226512, '9'),
            ('PC2', 'Armored Griffin', 271234, '1'),
            ('PC2', 'Chaotic Æther', 226509, '1'),
            ('PC2', 'Stairs to Infinity', 226521, 'P1'),
            ('MMA', 'Thallid', 370352, '167'),
            ('VMA', 'Academy Elite', 382835, '55'),
        ]
        self.assertCountEqual(expected, set_card_mv_number)

    def test_update_models_without_online_only(self):
        # Execute
        mtgjson.update_models(self.session, self.mtg_data, False)
        self.session.commit()

        # Verify
        printings = self.session.query(models.CardPrinting).all()
        set_card_mv_number = [
            (p.set.code, p.card.name, p.multiverseid, p.set_number)
            for p in printings]
        expected = [
            ('LEA', 'Dark Ritual', 54, None),
            ('LEA', 'Air Elemental', 94, None),
            ('LEA', 'Forest', 288, None),
            ('LEA', 'Forest', 289, None),
            ('FEM', 'Thallid', 1924, None),
            ('FEM', 'Thallid', 1926, None),
            ('FEM', 'Thallid', 1927, None),
            ('FEM', 'Thallid', 1925, None),
            ('pMEI', 'Arena', 97042, '1'),
            ('ICE', 'Dark Ritual', 2444, None),
            ('ICE', 'Forest', 2746, None),
            ('ICE', 'Forest', 2747, None),
            ('ICE', 'Forest', 2748, None),
            ('ICE', 'Snow-Covered Forest', 2749, None),
            ('HML', 'Cemetery Gate', 2913, None),
            ('HML', 'Cemetery Gate', 2914, None),
            ('S00', 'Rhox', None, None),
            ('PLS', 'Ertai, the Corrupted', 25614, '107'),
            ('PLS', 'Ertai, the Corrupted', 29292, '★107'),
            ('pMGD', 'Black Sun\'s Zenith', None, '7'),
            ('HOP', 'Academy at Tolaria West', 198073, '1'),
            ('HOP', 'Akroma\'s Vengeance', 205366, '1'),
            ('HOP', 'Dark Ritual', 205422, '24'),
            ('ARC', 'All in Good Time', 212648, '1'),
            ('ARC', 'Leonin Abunas', 220527, '1'),
            ('ISD', 'Abattoir Ghoul', 222911, '85'),
            ('ISD', 'Delver of Secrets', 226749, '51a'),
            ('ISD', 'Insectile Aberration', 226755, '51b'),
            ('ISD', 'Forest', 245247, '262'),
            ('ISD', 'Forest', 245248, '263'),
            ('ISD', 'Forest', 245246, '264'),
            ('PC2', 'Akoum', 226512, '9'),
            ('PC2', 'Armored Griffin', 271234, '1'),
            ('PC2', 'Chaotic Æther', 226509, '1'),
            ('PC2', 'Stairs to Infinity', 226521, 'P1'),
            ('MMA', 'Thallid', 370352, '167'),
        ]
        self.assertCountEqual(expected, set_card_mv_number)
